part of pdf;

/// Represents a flexible grid that consists of columns and rows.
class PdfGrid extends PdfLayoutElement {
  /// Initializes a new instance of the [PdfGrid] class.
  PdfGrid() {
    _initialize();
  }

  //Fields
  PdfGridColumnCollection _columns;
  PdfGridRowCollection _rows;
  PdfGridStyle _style;
  PdfGridHeaderCollection _headers;

  /// Gets or sets a value indicating whether to repeat header.
  bool repeatHeader;
  _Size _gridSize;

  /// Gets or sets a value indicating whether to split or cut rows that overflow a page.
  bool allowRowBreakingAcrossPages;
  bool _isChildGrid;
  PdfGridCell _parentCell;
  double _initialWidth;
  PdfLayoutFormat _layoutFormat;
  bool _isComplete;
  bool _isWidthSet;
  int _parentCellIndex;
  bool _isDrawn;
  double _rowLayoutBoundswidth;
  List<int> _listOfNavigatePages;
  bool _isRearranged;
  _Rectangle _gridLocation;
  bool _isPageWidth;
  bool _isSingleGrid;
  bool _hasColumnSpan;
  bool _hasRowSpan;
  PdfFont _defaultFont;
  DataTable _dataSource;

  //Events
  /// The event raised on starting cell lay outing.
  PdfGridBeginCellLayoutCallback beginCellLayout;

  /// The event raised on finished cell layout.
  PdfGridEndCellLayoutCallback endCellLayout;

  //Properties
  /// Gets the column collection of the PdfGrid.
  PdfGridColumnCollection get columns {
    _columns ??= PdfGridColumnCollection(this);
    return _columns;
  }

  /// Gets the row collection from the PdfGrid.
  PdfGridRowCollection get rows {
    _rows ??= PdfGridRowCollection(this);
    return _rows;
  }

  /// Gets the headers collection from the PdfGrid.
  PdfGridHeaderCollection get headers {
    _headers ??= PdfGridHeaderCollection(this);
    return _headers;
  }

  /// Gets the data bind to the [PdfGrid] by associating it with an external data source.
  DataTable get dataSource => _dataSource;

  /// Sets the data bind to the [PdfGrid] by associating it with an external data source.
  set dataSource(DataTable value) {
    ArgumentError.checkNotNull(value);
    _dataSource = value;
    _setData(this, value);
  }

  /// Gets the grid style.
  PdfGridStyle get style {
    _style ??= PdfGridStyle();
    return _style;
  }

  /// Sets the grid style.
  set style(PdfGridStyle value) {
    _style = value;
  }

  _Size get _size {
    if (_gridSize == null || _gridSize == _Size.empty) {
      _gridSize = _measure();
    }
    return _gridSize;
  }

  /// Gets a value indicating whether the start cell layout event should be raised.
  bool get _raiseBeginCellLayout => beginCellLayout != null;

  /// Gets a value indicating whether the end cell layout event should be raised.
  bool get _raiseEndCellLayout => endCellLayout != null;

  //Implementation
  void _initialize() {
    _gridSize = _Size.empty;
    _isComplete = false;
    _isDrawn = false;
    _isSingleGrid = true;
    _isWidthSet = false;
    repeatHeader = false;
    allowRowBreakingAcrossPages = true;
    _gridLocation = _Rectangle.empty;
    _hasColumnSpan = false;
    _hasRowSpan = false;
    _isChildGrid = false;
    _isPageWidth = false;
    _isRearranged = false;
    _initialWidth = 0;
    _rowLayoutBoundswidth = 0;
    _listOfNavigatePages = <int>[];
    _parentCellIndex = 0;
    _defaultFont = PdfStandardFont(PdfFontFamily.helvetica, 8);
  }

  PdfGrid _setData(PdfGrid grid, DataTable source) {
    final List<DataColumn> dataColumns = source.columns;
    if (dataColumns.isNotEmpty) {
      grid.headers.clear();
      grid.rows._rows.clear();
      grid.rows._rows.length = 0;
      grid.columns.add(count: dataColumns.length);
      grid.headers.add(1);
      for (int i = 0; i < dataColumns.length; i++) {
        final Widget cell = dataColumns[i].label;
        if (cell is Text) {
          grid.headers[0].cells[i].value = cell.data;
        } else if (cell is DataTable) {
          grid.headers[0].cells[i].value = _setData(PdfGrid(), cell);
        }
      }
    }
    final List<DataRow> dataRows = source.rows;
    if (dataRows.isNotEmpty) {
      final int columnCount = dataRows[0].cells.length;
      if (grid.columns.count <= 0 && columnCount > 0) {
        grid.columns.add(count: columnCount);
      }
      for (int i = 0; i < dataRows.length; i++) {
        final DataRow dataRow = dataRows[i];
        final PdfGridRow gridRow = grid.rows.add();
        for (int j = 0; j < columnCount; j++) {
          final DataCell dataCell = dataRow.cells[j];
          final Widget cell = dataCell.child;
          if (cell is Text) {
            gridRow.cells[j].value = cell.data;
          } else if (cell is DataTable) {
            gridRow.cells[j].value = _setData(PdfGrid(), cell);
          }
        }
      }
    }
    return grid;
  }

  _Size _measure() {
    double height = 0;
    final double width = columns._width;
    for (int i = 0; i < headers.count; i++) {
      height += headers[i].height;
    }
    for (int i = 0; i < rows.count; i++) {
      height += rows[i].height;
    }
    return _Size(width, height);
  }

  void _setSpan() {
    int colSpan, rowSpan = 1;
    int currentCellIndex, currentRowIndex = 0;
    int maxSpan = 0;
    for (int i = 0; i < headers.count; i++) {
      final PdfGridRow row = headers[i];
      maxSpan = 0;
      for (int j = 0; j < row.cells.count; j++) {
        final PdfGridCell cell = row.cells[j];
        maxSpan = max(maxSpan, cell.rowSpan);
        if (!cell._isCellMergeContinue &&
            !cell._isRowMergeContinue &&
            (cell.columnSpan > 1 || cell.rowSpan > 1)) {
          if (cell.columnSpan + j > row.cells.count) {
            throw ArgumentError.value(
                'Invalid span specified at row $j column $i');
          }
          if (cell.rowSpan + i > headers.count) {
            throw ArgumentError.value(
                'Invalid span specified at row $j column $i');
          }
          if (cell.columnSpan > 1 && cell.rowSpan > 1) {
            colSpan = cell.columnSpan;
            rowSpan = cell.rowSpan;
            currentCellIndex = j;
            currentRowIndex = i;
            while (colSpan > 1) {
              currentCellIndex++;
              row.cells[currentCellIndex]._isCellMergeContinue = true;
              row.cells[currentCellIndex]._isRowMergeContinue = true;
              row.cells[currentCellIndex].rowSpan = rowSpan;
              colSpan--;
            }
            currentCellIndex = j;
            colSpan = cell.columnSpan;
            while (rowSpan > 1) {
              currentRowIndex++;
              headers[currentRowIndex].cells[j]._isRowMergeContinue = true;
              headers[currentRowIndex]
                  .cells[currentCellIndex]
                  ._isRowMergeContinue = true;
              rowSpan--;
              while (colSpan > 1) {
                currentCellIndex++;
                headers[currentRowIndex]
                    .cells[currentCellIndex]
                    ._isCellMergeContinue = true;
                headers[currentRowIndex]
                    .cells[currentCellIndex]
                    ._isRowMergeContinue = true;
                colSpan--;
              }
              colSpan = cell.columnSpan;
              currentCellIndex = j;
            }
          } else if (cell.columnSpan > 1 && cell.rowSpan == 1) {
            colSpan = cell.columnSpan;
            currentCellIndex = j;
            while (colSpan > 1) {
              currentCellIndex++;
              row.cells[currentCellIndex]._isCellMergeContinue = true;
              colSpan--;
            }
          } else if (cell.columnSpan == 1 && cell.rowSpan > 1) {
            rowSpan = cell.rowSpan;
            currentRowIndex = i;
            while (rowSpan > 1) {
              currentRowIndex++;
              headers[currentRowIndex].cells[j]._isRowMergeContinue = true;
              rowSpan--;
            }
          }
        }
      }
      row._maximumRowSpan = maxSpan;
    }
    colSpan = rowSpan = 1;
    currentCellIndex = currentRowIndex = 0;
    if (_hasColumnSpan || _hasRowSpan) {
      for (int i = 0; i < rows.count; i++) {
        final PdfGridRow row = rows[i];
        maxSpan = 0;
        for (int j = 0; j < row.cells.count; j++) {
          final PdfGridCell cell = row.cells[j];
          maxSpan = max(maxSpan, cell.rowSpan);
          if (!cell._isCellMergeContinue &&
              !cell._isRowMergeContinue &&
              (cell.columnSpan > 1 || cell.rowSpan > 1)) {
            if (cell.columnSpan + j > row.cells.count) {
              throw ArgumentError.value(
                  'Invalid span specified at row $j column $i');
            }
            if (cell.rowSpan + i > rows.count) {
              throw ArgumentError.value(
                  'Invalid span specified at row $j column $i');
            }
            if (cell.columnSpan > 1 && cell.rowSpan > 1) {
              colSpan = cell.columnSpan;
              rowSpan = cell.rowSpan;
              currentCellIndex = j;
              currentRowIndex = i;
              while (colSpan > 1) {
                currentCellIndex++;
                row.cells[currentCellIndex]._isCellMergeContinue = true;
                row.cells[currentCellIndex]._isRowMergeContinue = true;
                colSpan--;
              }
              currentCellIndex = j;
              colSpan = cell.columnSpan;
              while (rowSpan > 1) {
                currentRowIndex++;
                rows[currentRowIndex].cells[j]._isRowMergeContinue = true;
                rows[currentRowIndex]
                    .cells[currentCellIndex]
                    ._isRowMergeContinue = true;
                rowSpan--;
                while (colSpan > 1) {
                  currentCellIndex++;
                  rows[currentRowIndex]
                      ._cells[currentCellIndex]
                      ._isCellMergeContinue = true;
                  rows[currentRowIndex]
                      ._cells[currentCellIndex]
                      ._isRowMergeContinue = true;
                  colSpan--;
                }
                colSpan = cell.columnSpan;
                currentCellIndex = j;
              }
            } else if (cell.columnSpan > 1 && cell.rowSpan == 1) {
              colSpan = cell.columnSpan;
              currentCellIndex = j;
              while (colSpan > 1) {
                currentCellIndex++;
                row.cells[currentCellIndex]._isCellMergeContinue = true;
                colSpan--;
              }
            } else if (cell.columnSpan == 1 && cell.rowSpan > 1) {
              rowSpan = cell.rowSpan;
              currentRowIndex = i;
              while (rowSpan > 1) {
                currentRowIndex++;
                rows[currentRowIndex].cells[j]._isRowMergeContinue = true;
                rowSpan--;
              }
            }
          }
        }
        row._maximumRowSpan = maxSpan;
      }
    }
  }

  void _measureColumnsWidth([_Rectangle bounds]) {
    if (bounds == null) {
      List<double> widths = List<double>.filled(columns.count, 0);
      double cellWidth = 0;
      if (headers.count > 0) {
        for (int i = 0; i < headers[0].cells.count; i++) {
          for (int j = 0; j < headers.count; j++) {
            cellWidth = max(
                cellWidth,
                _initialWidth > 0.0
                    ? min(_initialWidth, headers[j].cells[i].width)
                    : headers[j].cells[i].width);
          }
          widths[i] = cellWidth.toDouble();
        }
      }
      cellWidth = 0;
      for (int i = 0; i < columns.count; i++) {
        for (int j = 0; j < rows.count; j++) {
          final bool isGrid = rows[j].cells[i].value != null &&
              rows[j].cells[i].value is PdfGrid;
          if ((rows[j].cells[i].columnSpan == 1 &&
                  !rows[j].cells[i]._isCellMergeContinue) ||
              isGrid) {
            if (isGrid &&
                !rows[j]._grid.style.allowHorizontalOverflow &&
                _initialWidth != 0) {
              (rows[j].cells[i].value as PdfGrid)._initialWidth =
                  _initialWidth -
                      (rows[j]._grid.style.cellPadding.right +
                          rows[j]._grid.style.cellPadding.left +
                          rows[j].cells[i].style.borders.left.width / 2 +
                          _gridLocation.x);
            }
            cellWidth = max(
                widths[i],
                max(
                    cellWidth,
                    _initialWidth > 0.0
                        ? min(_initialWidth, rows[j].cells[i].width)
                        : rows[j].cells[i].width));
            cellWidth = max(columns[i].width, cellWidth);
          }
        }
        if (rows.count != 0) {
          widths[i] = cellWidth.toDouble();
        }
        cellWidth = 0;
      }
      for (int i = 0; i < rows.count; i++) {
        for (int j = 0; j < columns.count; j++) {
          if (rows[i].cells[j].columnSpan > 1) {
            double totalWidth = widths[j];
            for (int k = 1; k < rows[i].cells[j].columnSpan; k++) {
              totalWidth += widths[j + k];
            }
            if (rows[i].cells[j].width > totalWidth) {
              double extendedWidth = rows[i].cells[j].width - totalWidth;
              extendedWidth = extendedWidth / rows[i].cells[j].columnSpan;
              for (int k = j; k < j + rows[i].cells[j].columnSpan; k++) {
                widths[k] += extendedWidth;
              }
            }
          }
        }
      }
      if (_isChildGrid && _initialWidth != 0) {
        widths = columns._getDefaultWidths(_initialWidth);
      }
      for (int i = 0; i < columns.count; i++) {
        if (columns[i].width < 0) {
          columns[i]._width = widths[i];
        } else if (columns[i].width > 0 && !columns[i]._isCustomWidth) {
          columns[i]._width = widths[i];
        }
      }
    } else {
      List<double> widths = columns._getDefaultWidths(bounds.width - bounds.x);
      for (int i = 0; i < columns.count; i++) {
        if (columns[i].width < 0)
          columns[i]._width = widths[i];
        else if (columns[i].width > 0 &&
            !columns[i]._isCustomWidth &&
            widths[i] > 0 &&
            _isComplete) {
          columns[i]._width = widths[i];
        }
      }
      if (_parentCell != null &&
          (!style.allowHorizontalOverflow) &&
          (!_parentCell._row._grid.style.allowHorizontalOverflow)) {
        double padding = 0;
        double columnWidth = 0;
        int columnCount = columns.count;
        if (_parentCell.style.cellPadding != null) {
          padding += _parentCell.style.cellPadding.left +
              _parentCell.style.cellPadding.right;
        } else if (_parentCell._row._grid.style.cellPadding != null) {
          padding += _parentCell._row._grid.style.cellPadding.left +
              _parentCell._row._grid.style.cellPadding.right;
        }
        for (int i = 0; i < _parentCell.columnSpan; i++) {
          columnWidth +=
              _parentCell._row._grid.columns[_parentCellIndex + i].width;
        }
        for (int i = 0; i < columns.count; i++) {
          if (_columns[i].width > 0 && _columns[i]._isCustomWidth) {
            columnWidth -= _columns[i].width;
            columnCount--;
          }
        }
        if (columnWidth > padding) {
          final double childGridColumnWidth =
              (columnWidth - padding) / columnCount;
          if (_parentCell != null &&
              _parentCell.stringFormat.alignment != PdfTextAlignment.right) {
            for (int j = 0; j < columns.count; j++) {
              if (!columns[j]._isCustomWidth)
                columns[j]._width = childGridColumnWidth;
            }
          }
        }
      }
      if (_parentCell != null && _parentCell._row._getWidth() > 0) {
        if (_isChildGrid && _size.width > _parentCell._row._getWidth()) {
          widths = columns._getDefaultWidths(bounds.width);
          for (int i = 0; i < columns.count; i++) {
            columns[i].width = widths[i];
          }
        }
      }
    }
  }

  void _onBeginCellLayout(PdfGridBeginCellLayoutArgs args) {
    if (_raiseBeginCellLayout) {
      beginCellLayout(this, args);
    }
  }

  void _onEndCellLayout(PdfGridEndCellLayoutArgs args) {
    if (_raiseEndCellLayout) {
      endCellLayout(this, args);
    }
  }

  @override

  /// Draws the [PdfGrid]
  PdfLayoutResult draw(
      {Rect bounds,
      PdfLayoutFormat format,
      PdfGraphics graphics,
      PdfPage page}) {
    ArgumentError.checkNotNull(bounds);
    _initialWidth = bounds.width == 0
        ? page != null ? page.getClientSize().width : graphics.clientSize.width
        : bounds.width;
    _isWidthSet = true;
    if (graphics != null) {
      _drawInternal(graphics, _Rectangle.fromRect(bounds));
    } else if (page != null) {
      final PdfLayoutResult result =
          super.draw(page: page, bounds: bounds, format: format);
      _isComplete = true;
      return result;
    }
    return null;
  }

  @override
  void _drawInternal(PdfGraphics graphics, _Rectangle bounds) {
    _setSpan();
    final _PdfGridLayouter layouter = _PdfGridLayouter(this);
    layouter.layout(graphics, bounds);
    _isComplete = true;
  }

  @override
  PdfLayoutResult _layout(_PdfLayoutParams param) {
    if (param.bounds.width < 0) {
      ArgumentError.value('width', 'width', 'out of range');
    }
    if (rows.count != 0) {
      final PdfBorders borders = rows[0].cells[0].style.borders;
      if (borders != null && borders.left != null && borders.left.width != 1) {
        final double x = borders.left.width / 2;
        final double y = borders.top.width / 2;
        if (param.bounds.x == PdfBorders.defaultBorder.right.width / 2 &&
            param.bounds.y == PdfBorders.defaultBorder.right.width / 2) {
          param.bounds = _Rectangle(x, y, _size.width, _size.height);
        }
      }
    }
    _setSpan();
    _layoutFormat = param.format;
    _gridLocation = param.bounds;
    return _PdfGridLayouter(this)._layoutInternal(param);
  }
}

/// Delegate for handling StartCellLayoutEvent.
typedef PdfGridBeginCellLayoutCallback = void Function(
    Object sender, PdfGridBeginCellLayoutArgs args);

/// Delegate for handling EndCellLayoutEvent.
typedef PdfGridEndCellLayoutCallback = void Function(
    Object sender, PdfGridEndCellLayoutArgs args);

/// Represents arguments of StartCellLayout Event.
class PdfGridBeginCellLayoutArgs extends GridCellLayoutArgs {
  //Constructor
  PdfGridBeginCellLayoutArgs._(
      PdfGraphics graphics,
      int rowIndex,
      int cellInder,
      _Rectangle bounds,
      String value,
      PdfGridCellStyle style,
      bool isHeaderRow)
      : super._(graphics, rowIndex, cellInder, bounds, value, isHeaderRow) {
    if (style != null) {
      this.style = style;
    }
    skip = false;
  }
  //Fields
  /// PDF grid cell style
  PdfGridCellStyle style;

  /// A value that indicates whether the cell is drawn or not in the PDF document.
  bool skip;
}

/// Represents arguments of EndCellLayout Event.
class PdfGridEndCellLayoutArgs extends GridCellLayoutArgs {
  //Constructor
  PdfGridEndCellLayoutArgs._(PdfGraphics graphics, int rowIndex, int cellInder,
      _Rectangle bounds, String value, PdfGridCellStyle style, bool isHeaderRow)
      : super._(graphics, rowIndex, cellInder, bounds, value, isHeaderRow) {
    if (style != null) {
      this.style = style;
    }
  }
  //Fields
  /// PDF grid cell style
  PdfGridCellStyle style;
}

/// Represents the abstract class of the [GridCellLayoutArgs].
abstract class GridCellLayoutArgs {
  //Constructors
  /// Initializes a new instance of the [StartCellLayoutArgs] class.
  GridCellLayoutArgs._(PdfGraphics graphics, int rowIndex, int cellIndex,
      _Rectangle bounds, String value, bool isHeaderRow) {
    ArgumentError.checkNotNull(graphics);
    _rowIndex = rowIndex;
    _cellIndex = cellIndex;
    _value = value;
    _bounds = bounds.rect;
    _graphics = graphics;
    _isHeaderRow = isHeaderRow;
  }

  //Fields
  int _rowIndex;
  int _cellIndex;
  String _value;
  Rect _bounds;
  PdfGraphics _graphics;
  bool _isHeaderRow;

  //Properties
  /// Gets the index of the row.
  int get rowIndex => _rowIndex;

  /// Gets the index of the cell.
  int get cellIndex => _cellIndex;

  /// Gets the value.
  String get value => _value;

  /// Gets the bounds of the cell.
  Rect get bounds => _bounds;

  /// Gets the graphics, on which the cell should be drawn.
  PdfGraphics get graphics => _graphics;

  /// Gets the type of Grid row.
  bool get isHeaderRow => _isHeaderRow;
}

/// Arguments of BeginPageLayoutEvent.
class PdfGridBeginPageLayoutArgs extends BeginPageLayoutArgs {
  //Constructor
  PdfGridBeginPageLayoutArgs._(Rect bounds, PdfPage page, int startRow)
      : super(bounds, page) {
    startRowIndex = startRow != null ? startRow : 0;
  }

  //Fields
  /// Gets the start row index.
  int startRowIndex;
}

/// Arguments of EndPageLayoutEvent.
class PdfGridEndPageLayoutArgs extends EndPageLayoutArgs {
  //Constructor
  PdfGridEndPageLayoutArgs._(PdfLayoutResult result) : super(result);
}
